from ._dataset import DataSet
import logging
from typing import Any


class Write(DataSet):
    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        self._logger = logging.getLogger(self.__class__.__name__)
        self._render()



from pydantic import BaseModel, Field
from enum import Enum
from typing import Union, Dict, Any
import os


class SparkLoggingLevel(Enum):
    INFO = "INFO"
    DEBUG = "DEBUG"
    WARNING = "WARNING"
    ERROR = "ERROR"


class SparkConfig(BaseModel):
    logging_level: SparkLoggingLevel = Field(default=SparkLoggingLevel.INFO)
    config: Dict[str, Union[str, bool]] = Field(
        default={
            "spark.master": "local",
            "spark.sql.catalog.spark_catalog": "org.apache.spark.sql.delta.catalog.DeltaCatalog",
            "spark.databricks.delta.merge.repartitionBeforeWrite.enabled": True,
            "spark.sql.catalog.spark_catalog": "org.apache.spark.sql.delta.catalog.DeltaCatalog",
            "spark.sql.catalogImplementation": "hive",
        }
    )


class Project(BaseModel):
    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        self.sql = os.path.join(self.config_path, self.sql.replace("./", ""))
        self.pipelines = os.path.join(
            self.config_path, self.pipelines.replace("./", "")
        )
        self.databricks_notebooks = os.path.join(
            self.config_path, self.databricks_notebooks.replace("./", "")
        )
        self.databricks_workflows = os.path.join(
            self.config_path, self.databricks_workflows.replace("./", "")
        )
        self.databricks_queries = os.path.join(
            self.config_path, self.databricks_queries.replace("./", "")
        )

    config_path: str = Field(...)
    name: str = Field(...)
    sql: str = Field(default="./sql")
    pipelines: str = Field(default="./pipelines")
    databricks_notebooks: str = Field(default="./databricks/notebooks")
    databricks_workflows: str = Field(default="./databricks/workflows")
    databricks_queries: str = Field(default="./databricks/queries")

    spark: SparkConfig = Field(default=SparkConfig())



from ._deltalake import DeltaLake
from ._read import Read
from ._write import Write
from ._dataset import DataSet, Table
from ._deltalake import DeltaLakeTable
from ._read import ReadTable
import logging
from .dataset_type import DataSetType, TableType
from .._timeslice import Timeslice
from typing import Union


class _DatasetFactory:
    _TIMESLICE = "timeslice"
    _CONFIG_PATH = "config_path"

    def __init__(self) -> None:
        self._logger = logging.getLogger(self.__class__.__name__)
        self._dataset = {}
        self._table = {}

    def register_dataset_type(self, io_type: DataSetType, dataset_type: type):
        self._logger.debug(f"Register dataset type {dataset_type} as {type}")
        self._dataset[io_type] = dataset_type

    def register_table_type(self, io_type: TableType, table_type: type):
        self._logger.debug(f"Register table type {table_type} as {type}")
        self._table[io_type] = table_type

    def _get_dataset_type(
        self,
        table_type: TableType,
        dataset_config: dict,
    ) -> DataSet:
        self._logger.debug(f"Get DataSetType {table_type.value} from dataset factory")

        dataset_type: DataSetType = DataSetType(table_type.value)
        dataset_class = self._dataset.get(dataset_type)

        if not dataset_class:
            self._logger.debug(
                f"DataSetType {dataset_type.name} not registered in the dataset factory dataset"
            )
            raise ValueError(dataset_type)

        return dataset_class(
            **dataset_config,
        )

    def get_data_set(
        self, config: dict, dataset: Union[Table, dict], timeslice: Timeslice
    ):
        if isinstance(dataset, dict):
            for name, table_obj in dataset.items():
                stage_config = self._get_stage_table_config(
                    config, table_obj, timeslice
                )
                dataset[name] = self._get_dataset_type(
                    table_obj.table_type, stage_config
                )

        elif isinstance(dataset, Table):
            stage_config = self._get_stage_table_config(config, dataset, timeslice)
            dataset = self._get_dataset_type(dataset.table_type, stage_config)

        return dataset

    def get_table(self, table_type: TableType, table_config: dict):
        self._logger.debug(f"Get {table_type.name} from factory dataset")
        table_class = self._table.get(table_type)

        if not table_class:
            self._logger.debug(
                f"DataSetType {table_type.name} not registered in the dataset factory dataset"
            )
            raise ValueError(table_type)

        return table_class(
            **table_config,
        )

    def _get_stage_table_config(self, config: dict, table: Table, timeslice: Timeslice):
        """
        injects the table and ancillary attributes into the configuration
        for the typed data set object.
        """
        stage_config = config[table.stage.name][table.table_type.value]
        try:
            stage_config = config[table.stage.name][table.table_type.value]
        except Exception:
            raise Exception(
                f"""Table stage and type not found in configuration.
                Table: {table.stage.name}.{table.table_type.value}
                not int config:
                {config}."""
            )
        stage_config[self._TIMESLICE] = timeslice
        stage_config[self._CONFIG_PATH] = config[self._CONFIG_PATH]
        table_config = table.dict()
        stage_config = {**stage_config, **table_config}

        return stage_config


factory = _DatasetFactory()
factory.register_dataset_type(DataSetType.Read, Read)
factory.register_dataset_type(DataSetType.DeltaLake, DeltaLake)
factory.register_dataset_type(DataSetType.Write, Write)

factory.register_table_type(TableType.ReadTable, ReadTable)
factory.register_table_type(TableType.DeltaLakeTable, DeltaLakeTable)



from pydantic import BaseModel, Field
from typing import Union, Any, Dict, List
from ._stage_type import StageType
import fnmatch
from ._table_mapping import TableMapping
from .dataset.dataset_type import TableType
from .dataset import dataset_factory
from .dataset._dataset import Table


_INDEX_WILDCARD = "*"


class Tables(BaseModel):
    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        self._load_index()

    table_data: dict = Field(...)
    tables_index: Dict[str, Table] = Field(default={})
    delta_properties: Dict[str, str] = Field(default=None)
    config_path: str = Field(...)

    @classmethod
    def get_index(
        cls,
        stage: Union[StageType, str] = _INDEX_WILDCARD,
        database=_INDEX_WILDCARD,
        table=_INDEX_WILDCARD,
    ):
        if isinstance(stage, StageType):
            return f"{stage.name}.{database}.{table}"
        else:
            return f"{stage}.{database}.{table}"

    @classmethod
    def parse_index(
        cls,
        index: str,
    ):
        try:
            parts = index.split(".")
            stage = StageType[parts[0]]
            database = parts[1]
            table = parts[2]
        except Exception as e:
            raise Exception(
                f"attempted to parse an invalid index {index}. It must be of the form 'stage.database.table'"
            ) from e

        return stage, database, table

    def _parse_table_delta_lake_properties(
        self,
        table_type: TableType,
        table_details: dict,
        merge_properties: dict = {},
    ):
        """
        Parse the delta properties from the tables details if there are any. Merge the into
        a set of existing properties with the table detail properties taking precedence.
        """
        if table_type is TableType.DeltaLakeTable:
            delta_properties = table_details.get("delta_properties", {})
            # delta_properties = stage_delta_properties | delta_properties
            delta_properties = {
                **merge_properties,
                **delta_properties,
            }
            return delta_properties
        else:
            return None

    def _parse_stage_delta_lake_properties(
        self, table_type: TableType, database_tables: dict
    ):
        """Parse delta lake properties from the database table details and remove it"""
        stage_delta_properties = None
        if table_type is TableType.DeltaLakeTable:
            stage_delta_properties = database_tables.get("delta_properties", {})
            if stage_delta_properties:
                del database_tables["delta_properties"]
        return stage_delta_properties

    def _parse_stage_data(self, stage: StageType, stage_data: dict):
        """
        loop through the database tables, create the table
        and add it to the class index.
        """
        for table_type, database_tables in stage_data.items():
            table_type = TableType(table_type)
            stage_delta_properties = self._parse_stage_delta_lake_properties(
                table_type, database_tables
            )
            for database, tables in database_tables.items():
                for table, table_details in tables.items():
                    # flatten the config structure for a table
                    if not table_details:
                        table_details = {}

                    delta_properties = self._parse_table_delta_lake_properties(
                        table_type, table_details, stage_delta_properties
                    )
                    if delta_properties:
                        table_details["delta_properties"] = delta_properties
                    table_details["table"] = table
                    table_details["database"] = database
                    table_details["stage"] = stage
                    table_details["table_type"] = table_type

                    # create a table object

                    # table = Table(**table_details)
                    table = dataset_factory.get_table(table_type, table_details)

                    # index the table object
                    index = f"{stage.name}.{database}.{table.table}"
                    self.tables_index[index] = table

    def _load_index(self):
        """
        Parse through the table definitions dictionary and deserialize it
        into Table objects. The table object are then place in a dictionary for easy
        lookup with a key = stage.database.table and the value being the table
        object it self. This dictionary index is held on self.tables_index
        """
        for stage in StageType:
            stage_data = self.table_data.get(stage)
            if stage_data:
                self._parse_stage_data(stage, stage_data)

    def lookup_table(
        self,
        stage: Union[StageType, str] = _INDEX_WILDCARD,
        database=_INDEX_WILDCARD,
        table=_INDEX_WILDCARD,
        first_match: bool = True,
        **kwargs,
    ):
        index = Tables.get_index(stage, database, table)
        matches = fnmatch.filter(list(self.tables_index.keys()), index)

        if not matches:
            raise Exception(f"index {index} not found in tables_index")

        def match_property(
            table: Table, properties: Dict[str, Any], matches: List[str]
        ):
            for p, v in properties.items():
                if (
                    isinstance(table.custom_properties, dict)
                    and table.custom_properties.get(p) == v
                ):
                    return True
                else:
                    index = Tables.get_index(table.stage, table.database, table.name)
                    if index in matches:
                        matches.remove(
                            Tables.get_index(table.stage, table.database, table.name)
                        )
                    return False

        tables_index = dict(self.tables_index)
        if kwargs:
            tables_index = {
                k: v
                for k, v in self.tables_index.items()
                if match_property(v, kwargs, matches)
            }

        if first_match:
            matches = matches[0]
            table = tables_index[matches]
            return table
        else:
            tables = [tables_index[i] for i in matches]
            return tables

    def get_table_mapping(
        self, stage: StageType, table=_INDEX_WILDCARD, database=_INDEX_WILDCARD
    ):
        destination = self.lookup_table(
            stage=stage, database=database, table=table, first_match=True
        )
        source = {}

        try:
            for index in destination.depends_on:
                do_stage, do_database, do_table = Tables.parse_index(index)
                table = self.lookup_table(
                    stage=do_stage,
                    table=do_table,
                    database=do_database,
                    first_match=True,
                )
                source[table.table] = table
        except Exception as e:
            raise Exception(f"Error looking up dependencies for table {table}") from e

        if len(list(source.values())) == 1:
            source = list(source.values())[0]

        return TableMapping(source=source, destination=destination)



import logging
import logging.config
import yaml
import os
from yaml import YAMLError
from ._utils import YETL_CONFIG


def configure_logging(project: str = None, config_path=None):
    """Returns as of a named logger based on a yaml logging configuration file

    The configuration file is called logging.yaml. It's a directory
    location is taken from an envrionment variables called DATAPREPHOME.
    If this does not exist then it is defaulted to ./config/
    """
    log_config_file = config_path
    if not log_config_file:
        log_config_file = os.getenv(YETL_CONFIG, "./config")

    if project:
        project_log_config_file = os.path.join(log_config_file, project)

    project_log_config_file = f"{project_log_config_file}/logging.yaml"
    project_log_config_file = os.path.abspath(project_log_config_file)

    log_config_file = f"{log_config_file}/logging.yaml"
    log_config_file = os.path.abspath(log_config_file)

    if os.path.exists(project_log_config_file):
        log_config_file = project_log_config_file

    # check that it exists
    if not os.path.exists(log_config_file):
        msg = f"Config logging file path does not exist {log_config_file}"
        raise Exception(msg)

    # load the logging configuration into the logger
    with open(log_config_file, "r") as f:
        try:
            config = yaml.safe_load(f.read())
            logging.config.dictConfig(config)

            # if it errors because of invalid yaml format then
            # provide details so the users can easily find and correct
            # if it's a different exception just let it raise
        except YAMLError as e:
            location = ""
            if hasattr(e, "problem_mark"):
                mark = e.problem_mark
                location = f"Error position ({mark.line}, {mark.column})"

            if hasattr(e, "problem"):
                problem = f"{e.problem}."

            raise Exception(
                f"Invalid yaml format in {log_config_file}. {problem} {location}"
            )
